# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2015 Yann Lanthony
# Copyright (c) 2017-2018 Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (See LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""Test qtsass conformers."""

from __future__ import absolute_import

# Standard library imports
from textwrap import dedent
import unittest

# Local imports
from qtsass.conformers import (
    NotConformer,
    QLinearGradientConformer,
    QRadialGradientConformer,
)


class TestNotConformer(unittest.TestCase):

    qss_str = 'QAbstractItemView::item:!active'
    css_str = 'QAbstractItemView::item:_qnot_active'

    def test_conform_to_scss(self):
        """NotConformer qss to scss."""

        c = NotConformer()
        self.assertEqual(c.to_scss(self.qss_str), self.css_str)

    def test_conform_to_qss(self):
        """NotConformer css to qss."""

        c = NotConformer()
        self.assertEqual(c.to_qss(self.css_str), self.qss_str)

    def test_round_trip(self):
        """NotConformer roundtrip."""

        c = NotConformer()
        conformed_css = c.to_scss(self.qss_str)
        self.assertEqual(c.to_qss(conformed_css), self.qss_str)


class TestQLinearGradientConformer(unittest.TestCase):

    css_vars_str = 'qlineargradient($x1, $y1, $x2, $y2, (0 $red, 1 $blue))'
    qss_vars_str = (
        'qlineargradient(x1:$x1, x2:$x2, y1:$y1, y2:$y2'
        'stop: 0 $red, stop: 1 $blue)'
    )

    css_nostops_str = 'qlineargradient(0, 0, 0, 0)'
    qss_nostops_str = 'qlineargradient(x1: 0, y1: 0, x2: 0, y2: 0)'

    css_str = 'qlineargradient(0, 0, 0, 0, (0 red, 1 blue))'
    qss_singleline_str = (
        'qlineargradient(x1: 0, y1: 0, x2: 0, y2: 0, '
        'stop: 0 red, stop: 1 blue)'
    )
    qss_multiline_str = dedent("""
    qlineargradient(
        x1: 0,
        y1: 0,
        x2: 0,
        y2: 0,
        stop: 0 red,
        stop: 1 blue
    )
    """).strip()
    qss_weird_whitespace_str = (
        'qlineargradient( x1: 0, y1:0, x2: 0, y2:0, '
        '   stop:0 red, stop: 1 blue )'
    )

    css_rgba_str = (
        'qlineargradient(0, 0, 0, 0, '
        '(0 rgba(0, 1, 2, 30%), 0.99 rgba(7, 8, 9, 100%)))'
    )
    qss_rgba_str = (
        'qlineargradient(x1: 0, y1: 0, x2: 0, y2: 0, '
        'stop: 0 rgba(0, 1, 2, 30%), stop: 0.99 rgba(7, 8, 9, 100%))'
    )

    css_incomplete_coords_str = (
        'qlineargradient(0, 1, 0, 0, (0 red, 1 blue))'
    )

    qss_incomplete_coords_str = (
        'qlineargradient(y1:1, stop:0 red, stop: 1 blue)'
    )

    css_float_coords_str = (
        'qlineargradient(0, 0.75, 0, 0, (0 green, 1 pink))'
    )

    qss_float_coords_str = (
        'qlineargradient(y1:0.75, stop:0 green, stop: 1 pink)'
    )

    def test_does_not_affect_css_form(self):
        """QLinearGradientConformer no affect on css qlineargradient func."""

        c = QLinearGradientConformer()
        self.assertEqual(c.to_scss(self.css_str), self.css_str)
        self.assertEqual(c.to_qss(self.css_str), self.css_str)

    def test_conform_singleline_str(self):
        """QLinearGradientConformer singleline qss to scss."""

        c = QLinearGradientConformer()
        self.assertEqual(c.to_scss(self.qss_singleline_str), self.css_str)

    def test_conform_multiline_str(self):
        """QLinearGradientConformer multiline qss to scss."""

        c = QLinearGradientConformer()
        self.assertEqual(c.to_scss(self.qss_multiline_str), self.css_str)

    def test_conform_weird_whitespace_str(self):
        """QLinearGradientConformer weird whitespace qss to scss."""

        c = QLinearGradientConformer()
        self.assertEqual(c.to_scss(self.qss_weird_whitespace_str), self.css_str)

    def test_conform_nostops_str(self):
        """QLinearGradientConformer qss with no stops to scss."""

        c = QLinearGradientConformer()
        self.assertEqual(c.to_scss(self.qss_nostops_str), self.css_nostops_str)

    def test_conform_vars_str(self):
        """QLinearGradientConformer qss with vars to scss."""

        c = QLinearGradientConformer()
        self.assertEqual(c.to_scss(self.qss_vars_str), self.css_vars_str)

    def test_conform_rgba_str(self):
        """QLinearGradientConformer qss with rgba to scss."""

        c = QLinearGradientConformer()
        self.assertEqual(c.to_scss(self.qss_rgba_str), self.css_rgba_str)

    def test_incomplete_coords(self):
        """QLinearGradientConformer qss with not all 4 coordinates given."""

        c = QLinearGradientConformer()
        self.assertEqual(c.to_scss(self.qss_incomplete_coords_str),
                         self.css_incomplete_coords_str)

    def test_float_coords(self):
        c = QLinearGradientConformer()
        self.assertEqual(c.to_scss(self.qss_float_coords_str),
                         self.css_float_coords_str)


class TestQRadialGradientConformer(unittest.TestCase):

    css_vars_str = "qradialgradient('$spread', $cx, $cy, $radius, $fx, $fy, (0 $red, 1 $blue))"
    qss_vars_str = (
        'qradialgradient(spread:$spread, cx:$cx, cy:$cy, radius:$radius, fx:$fx, fy:$fy,'
        'stop: 0 $red, stop: 1 $blue)'
    )

    css_nostops_str = "qradialgradient('pad', 0, 0, 0, 0, 0)"
    qss_nostops_str = 'qradialgradient(spread: pad, cx: 0, cy: 0, fx: 0, fy: 0)'

    css_str = "qradialgradient('pad', 0, 0, 0, 0, 0, (0 red, 1 blue))"
    qss_singleline_str = (
        'qradialgradient(spread: pad, cx: 0, cy: 0, fx: 0, fy: 0, '
        'stop: 0 red, stop: 1 blue)'
    )
    qss_multiline_str = dedent("""
    qradialgradient(
        spread: pad,
        cx: 0,
        cy: 0,
        fx: 0,
        fy: 0,
        stop: 0 red,
        stop: 1 blue
    )
    """).strip()
    qss_weird_whitespace_str = (
        'qradialgradient( spread: pad, cx: 0, cy:0, fx: 0, fy:0, '
        '   stop:0 red, stop: 1 blue )'
    )

    css_rgba_str = (
        "qradialgradient('pad', 0, 0, 0, 0, 0, "
        "(0 rgba(0, 1, 2, 30%), 0.99 rgba(7, 8, 9, 100%)))"
    )
    qss_rgba_str = (
        'qradialgradient(spread: pad, cx: 0, cy: 0, fx: 0, fy: 0, '
        'stop: 0 rgba(0, 1, 2, 30%), stop: 0.99 rgba(7, 8, 9, 100%))'
    )

    css_incomplete_coords_str = (
        "qradialgradient('pad', 0, 1, 0, 0, 0, (0 red, 1 blue))"
    )

    qss_incomplete_coords_str = (
        'qradialgradient(spread:pad, cy:1, stop:0 red, stop: 1 blue)'
    )

    css_float_coords_str = (
        "qradialgradient('pad', 0, 0.75, 0, 0, 0, (0 green, 1 pink))"
    )

    qss_float_coords_str = (
        'qradialgradient(spread: pad, cy:0.75, stop:0 green, stop: 1 pink)'
    )

    def test_does_not_affect_css_form(self):
        """QRadialGradientConformer no affect on css qradialgradient func."""

        c = QRadialGradientConformer()
        self.assertEqual(c.to_scss(self.css_str), self.css_str)
        self.assertEqual(c.to_qss(self.css_str), self.css_str)

    def test_conform_singleline_str(self):
        """QRadialGradientConformer singleline qss to scss."""

        c = QRadialGradientConformer()
        self.assertEqual(c.to_scss(self.qss_singleline_str), self.css_str)

    def test_conform_multiline_str(self):
        """QRadialGradientConformer multiline qss to scss."""

        c = QRadialGradientConformer()
        self.assertEqual(c.to_scss(self.qss_multiline_str), self.css_str)

    def test_conform_weird_whitespace_str(self):
        """QRadialGradientConformer weird whitespace qss to scss."""

        c = QRadialGradientConformer()
        self.assertEqual(c.to_scss(self.qss_weird_whitespace_str), self.css_str)

    def test_conform_nostops_str(self):
        """QRadialGradientConformer qss with no stops to scss."""

        c = QRadialGradientConformer()
        self.assertEqual(c.to_scss(self.qss_nostops_str), self.css_nostops_str)

    def test_conform_vars_str(self):
        """QRadialGradientConformer qss with vars to scss."""

        c = QRadialGradientConformer()
        self.assertEqual(c.to_scss(self.qss_vars_str), self.css_vars_str)

    def test_conform_rgba_str(self):
        """QRadialGradientConformer qss with rgba to scss."""

        c = QRadialGradientConformer()
        self.assertEqual(c.to_scss(self.qss_rgba_str), self.css_rgba_str)

    def test_incomplete_coords(self):
        """QRadialGradientConformer qss with not all 4 coordinates given."""

        c = QRadialGradientConformer()
        self.assertEqual(c.to_scss(self.qss_incomplete_coords_str),
                         self.css_incomplete_coords_str)

    def test_float_coords(self):
        c = QRadialGradientConformer()
        self.assertEqual(c.to_scss(self.qss_float_coords_str),
                         self.css_float_coords_str)


if __name__ == "__main__":
    unittest.main(verbosity=2)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2015 Yann Lanthony
# Copyright (c) 2017-2018 Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (See LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""Test qtsass custom functions."""

from __future__ import absolute_import

# Standard library imports
import unittest

# Local imports
from qtsass.api import compile


class BaseCompileTest(unittest.TestCase):
    def compile_scss(self, string):
        # NOTE: revise for better future compatibility
        wstr = '*{{t: {0};}}'.format(string)
        res = compile(wstr)
        return res.replace('* {\n  t: ', '').replace('; }\n', '')


class TestRgbaFunc(BaseCompileTest):
    def test_rgba(self):
        self.assertEqual(
            self.compile_scss('rgba(0, 1, 2, 0.3)'),
            'rgba(0, 1, 2, 30%)'
        )

    def test_rgba_percentage_alpha(self):
        result = self.compile_scss('rgba(255, 0, 125, 75%)')
        self.assertEqual(result, 'rgba(255, 0, 125, 75%)')

    def test_rgba_8bit_int_alpha(self):
        for in_val, out_val in ((0, 0), (128, 50), (255, 100)):
            result = self.compile_scss('rgba(255, 0, 125, %i)' % in_val)
            self.assertEqual(result, 'rgba(255, 0, 125, %i%%)' % out_val)


class TestQLinearGradientFunc(BaseCompileTest):
    def test_color(self):
        self.assertEqual(
            self.compile_scss('qlineargradient(1, 2, 3, 4, (0 red, 1 blue))'),
            'qlineargradient(x1: 1.0, y1: 2.0, x2: 3.0, y2: 4.0, '
            'stop: 0.0 rgba(255, 0, 0, 100%), stop: 1.0 rgba(0, 0, 255, 100%))'
        )

    def test_rgba(self):
        self.assertEqual(
            self.compile_scss('qlineargradient(1, 2, 3, 4, (0 red, 0.2 rgba(5, 6, 7, 0.8)))'),
            'qlineargradient(x1: 1.0, y1: 2.0, x2: 3.0, y2: 4.0, '
            'stop: 0.0 rgba(255, 0, 0, 100%), stop: 0.2 rgba(5, 6, 7, 80%))'
        )


class TestQRadialGradientFunc(BaseCompileTest):
    def test_color(self):
        self.assertEqual(
            self.compile_scss('qradialgradient(pad, 1, 2, 1, 3, 4, (0 red, 1 blue))'),
            'qradialgradient(spread: pad, cx: 1.0, cy: 2.0, radius: 1.0, fx: 3.0, fy: 4.0, '
            'stop: 0.0 rgba(255, 0, 0, 100%), stop: 1.0 rgba(0, 0, 255, 100%))'
        )

    def test_rgba(self):
        self.assertEqual(
            self.compile_scss('qradialgradient(pad, 1, 2, 1, 3, 4, (0 red, 0.2 rgba(5, 6, 7, 0.8)))'),
            'qradialgradient(spread: pad, cx: 1.0, cy: 2.0, radius: 1.0, fx: 3.0, fy: 4.0, '
            'stop: 0.0 rgba(255, 0, 0, 100%), stop: 0.2 rgba(5, 6, 7, 80%))'
        )


if __name__ == "__main__":
    unittest.main(verbosity=2)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2015 Yann Lanthony
# Copyright (c) 2017-2018 Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (See LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""Test qtsass api."""

from __future__ import absolute_import

# Standard library imports
from os.path import exists
import logging

# Third party imports
import pytest
import sass

# Local imports
import qtsass

# Local imports
from . import EXAMPLES_DIR, PROJECT_DIR, example


COLORS_STR = """
QWidget {
    background: rgba(127, 127, 127, 100%);
    color: rgb(255, 255, 255);
}
"""
QLINEARGRADIENTS_STR = """
QWidget {
    background: qlineargradient(
        x1: 0,
        y1: 0,
        x2: 0,
        y2: 1,
        stop: 0.1 blue,
        stop: 0.8 green
    );
}
"""
QRADIANTGRADIENTS_STR = """
QWidget {
    background: qradialgradient(
        spread: repeat,
        cx: 0,
        cy: 0,
        fx: 0,
        fy: 1,
        stop: 0.1 blue,
        stop: 0.8 green
    );
}
"""
QNOT_STR = """
QLineEdit:!editable {
    background: white;
}
"""
IMPORT_STR = """
@import 'dummy';
"""
CUSTOM_BORDER_STR = """
QWidget {
    border: custom_border();
}
"""


def setup_module():
    qtsass.enable_logging(level=logging.DEBUG)


def teardown_module():
    qtsass.enable_logging(level=logging.WARNING)


def test_compile_strings():
    """compile various strings."""

    qtsass.compile(COLORS_STR)
    qtsass.compile(QLINEARGRADIENTS_STR)
    qtsass.compile(QRADIANTGRADIENTS_STR)
    qtsass.compile(QNOT_STR)


def test_compile_import_raises():
    """compile string with import raises."""

    with pytest.raises(sass.CompileError):
        qtsass.compile(IMPORT_STR)


def test_compile_import_with_include_paths():
    """compile string with include_paths"""

    qtsass.compile(IMPORT_STR, include_paths=[EXAMPLES_DIR])


def test_compile_raises_ValueError():
    """compile raises ValueError with invalid arguments"""

    # Pass invalid type to importers - must be sequence
    with pytest.raises(ValueError):
        qtsass.compile(COLORS_STR, importers=lambda x: None)

    # Pass invalid type to custom_functions
    with pytest.raises(ValueError):
        qtsass.compile(COLORS_STR, custom_functions=lambda x: None)


def test_compile_custom_function():
    """compile string with custom_functions"""

    custom_str = (
        'QWidget {\n'
        '    border: custom_border();\n'
        '}'
    )

    def custom_border():
        return '1px solid'

    css = qtsass.compile(custom_str, custom_functions=[custom_border])
    assert '1px solid' in css
    assert 'custom_border()' not in css


def test_compile_filename(tmpdir):
    """compile_filename simple."""

    output = tmpdir.join('dummy.css')
    qtsass.compile_filename(example('dummy.scss'), output.strpath)
    assert exists(output.strpath)


def test_compile_filename_no_save():
    """compile_filename simple."""

    qss = qtsass.compile_filename(example('dummy.scss'))
    assert isinstance(qss, str)


def test_compile_filename_imports(tmpdir):
    """compile_filename with imports."""

    output = tmpdir.join('dark.css')
    qtsass.compile_filename(example('complex', 'dark.scss'), output.strpath)
    assert exists(output.strpath)


def test_compile_filename_imports_no_save():
    """compile_filename with imports."""

    qss = qtsass.compile_filename(example('complex', 'dark.scss'))
    assert isinstance(qss, str)


def test_compile_dirname(tmpdir):
    """compile_dirname complex."""

    output = tmpdir.join('complex')
    qtsass.compile_dirname(example('complex'), output.strpath)
    assert exists(output.join('dark.css').strpath)
    assert exists(output.join('light.css').strpath)


def test_watch_raises_ValueError(tmpdir):
    """watch raises ValueError when source does not exist."""

    # Watch file does not raise
    _ = qtsass.watch(example('dummy.scss'), tmpdir.join('dummy.scss').strpath)

    # Watch dir does not raise
    _ = qtsass.watch(example('complex'), tmpdir.join('complex').strpath)

    with pytest.raises(ValueError):
        _ = qtsass.watch('does_not_exist', 'does_not_exist')



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2015 Yann Lanthony
# Copyright (c) 2017-2018 Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (See LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""Test qtsass cli."""

from __future__ import absolute_import

# Standard library imports
from collections import namedtuple
from os.path import basename, exists
from subprocess import PIPE, Popen
import time

# Local imports
from . import PROJECT_DIR, await_condition, example, touch


SLEEP_INTERVAL = 1
Result = namedtuple('Result', "code stdout stderr")


def indent(text, prefix='    '):
    """Like textwrap.indent"""

    return ''.join([prefix + line for line in text.splitlines(True)])


def invoke(args):
    """Invoke qtsass cli with specified args"""

    kwargs = dict(
        stdout=PIPE,
        stderr=PIPE,
        cwd=PROJECT_DIR
    )
    proc = Popen(['python', '-m', 'qtsass'] + args, **kwargs)
    return proc


def invoke_with_result(args):
    """Invoke qtsass cli and return a Result obj"""

    proc = invoke(args)
    out, err = proc.communicate()
    out = out.decode('ascii', errors="ignore")
    err = err.decode('ascii', errors="ignore")
    return Result(proc.returncode, out, err)


def kill(proc, timeout=1):
    """Kill a subprocess and return a Result obj"""

    proc.kill()
    out, err = proc.communicate()
    out = out.decode('ascii', errors="ignore")
    err = err.decode('ascii', errors="ignore")
    return Result(proc.returncode, out, err)


def format_result(result):
    """Format a subprocess Result obj"""

    out = [
        'Subprocess Report...',
        'Exit code: %s' % result.code,
    ]
    if result.stdout:
        out.append('stdout:')
        out.append(indent(result.stdout, '    '))
    if result.stderr:
        out.append('stderr:')
        out.append(indent(result.stderr, '    '))
    return '\n'.join(out)


def test_compile_dummy_to_stdout():
    """CLI compile dummy example to stdout."""

    args = [example('dummy.scss')]
    result = invoke_with_result(args)

    assert result.code == 0
    assert result.stdout


def test_compile_dummy_to_file(tmpdir):
    """CLI compile dummy example to file."""

    input = example('dummy.scss')
    output = tmpdir.join('dummy.css')
    args = [input, '-o', output.strpath]
    result = invoke_with_result(args)

    assert result.code == 0
    assert exists(output.strpath)


def test_watch_dummy(tmpdir):
    """CLI watch dummy example."""

    input = example('dummy.scss')
    output = tmpdir.join('dummy.css')
    args = [input, '-o', output.strpath, '-w']
    proc = invoke(args)

    # Wait for initial compile
    output_exists = lambda: exists(output.strpath)
    if not await_condition(output_exists):
        result = kill(proc)
        report = format_result(result)
        err = "Failed to compile dummy.scss\n"
        err += report
        assert False, report

    # Ensure subprocess is still alive
    assert proc.poll() is None

    # Touch input file, triggering a recompile
    created = output.mtime()
    file_modified = lambda: output.mtime() > created
    time.sleep(SLEEP_INTERVAL)
    touch(input)

    if not await_condition(file_modified):
        result = kill(proc)
        report = format_result(result)
        err = 'Modifying %s did not trigger recompile.\n' % basename(input)
        err += report
        assert False, err

    kill(proc)


def test_compile_complex(tmpdir):
    """CLI compile complex example."""

    input = example('complex')
    output = tmpdir.mkdir('output')
    args = [input, '-o', output.strpath]
    result = invoke_with_result(args)

    assert result.code == 0

    expected_files = [output.join('light.css'), output.join('dark.css')]
    for file in expected_files:
        assert exists(file.strpath)


def test_watch_complex(tmpdir):
    """CLI watch complex example."""

    input = example('complex')
    output = tmpdir.mkdir('output')
    args = [input, '-o', output.strpath, '-w']
    proc = invoke(args)

    expected_files = [output.join('light.css'), output.join('dark.css')]

    # Wait for initial compile
    files_created = lambda: all([exists(f.strpath) for f in expected_files])
    if not await_condition(files_created):
        result = kill(proc)
        report = format_result(result)
        err = 'All expected files have not been created...'
        err += report
        assert False, err

    # Ensure subprocess is still alive
    assert proc.poll() is None

    # Input files to touch
    input_full = example('complex', 'light.scss')
    input_partial = example('complex', '_base.scss')
    input_nested = example('complex', 'widgets', '_qwidget.scss')

    def touch_and_wait(input_file, timeout=2000):
        """Touch a file, triggering a recompile"""

        filename = basename(input_file)
        old_mtimes = [f.mtime() for f in expected_files]
        files_modified = lambda: all(
            [f.mtime() > old_mtimes[i] for i, f in enumerate(expected_files)]
        )
        time.sleep(SLEEP_INTERVAL)
        touch(input_file)

        if not await_condition(files_modified, timeout):
            result = kill(proc)
            report = format_result(result)
            err = 'Modifying %s did not trigger recompile.\n' % filename
            err += report
            for i, f in enumerate(expected_files):
                err += str(f) + '\n'
                err += str(old_mtimes[i]) + '\n'
                err += str(f.mtime()) + '\n'
                err += str(bool(f.mtime() > old_mtimes[i])) + '\n'
            assert False, err

        return True

    assert touch_and_wait(input_full)
    assert touch_and_wait(input_partial)
    assert touch_and_wait(input_nested)

    kill(proc)


def test_invalid_input():
    """CLI input is not a file or dir."""

    proc = invoke_with_result(['file_does_not_exist.scss'])
    assert proc.code == 1
    assert 'Error: input must be' in proc.stdout

    proc = invoke_with_result(['./dir/does/not/exist'])
    assert proc.code == 1
    assert 'Error: input must be' in proc.stdout


def test_dir_missing_output():
    """CLI dir missing output option"""

    proc = invoke_with_result([example('complex')])
    assert proc.code == 1
    assert 'Error: missing required option' in proc.stdout



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2015 Yann Lanthony
# Copyright (c) 2017-2018 Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (See LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""Test qtsass cli."""

from __future__ import absolute_import

# Standard library imports
from os.path import dirname, exists
import os
import shutil
import sys
import time

# Third party imports
from flaky import flaky
import pytest

# Local imports
#Local imports
from qtsass import compile_filename
from qtsass.watchers import PollingWatcher, QtWatcher
from qtsass.watchers.api import retry

# Local imports
from . import EXAMPLES_DIR, await_condition, example, touch


class CallCounter(object):

    def __init__(self):
        self.count = 0

    def __call__(self, *args, **kwargs):
        self.count += 1


@pytest.mark.parametrize(
    'Watcher', (PollingWatcher, QtWatcher),
)
@flaky(max_runs=3)
def test_watchers(Watcher, tmpdir):
    """Stress test Watcher implementations"""

    # Skip when QtWatcher is None - when Qt is not installed.
    if not Watcher:
        return

    watch_dir = tmpdir.join('src').strpath
    os.makedirs(watch_dir)
    shutil.copy2(example('dummy.scss'), watch_dir)
    input = tmpdir.join('src/dummy.scss').strpath
    output = tmpdir.join('build/dummy.css').strpath
    output_exists = lambda: exists(output)

    c = CallCounter()
    w = Watcher(
        watch_dir=watch_dir,
        compiler=compile_filename,
        args=(input, output),
    )
    w.connect(c)

    # Output should not yet exist
    assert not exists(output)

    w.start()

    touch(input)
    time.sleep(0.5)
    if not await_condition(output_exists):
        assert False, 'Output file not created...'

    # Removing the watch_dir should not kill the Watcher
    # simply stop dispatching callbacks
    shutil.rmtree(watch_dir)
    time.sleep(0.5)
    assert c.count == 1

    # Watcher should recover once the input file is there again
    os.makedirs(watch_dir)
    shutil.copy2(example('dummy.scss'), watch_dir)
    time.sleep(0.5)
    assert c.count == 2

    # Stop watcher
    w.stop()
    w.join()

    for _ in range(5):
        touch(input)

    # Count should not change
    assert c.count == 2


@pytest.mark.skipif(sys.platform.startswith('linux') or not QtWatcher,
                    reason="Fails on linux")
def test_qtwatcher(tmpdir):
    """Test QtWatcher implementation."""
    # Constructing a QApplication will cause the QtWatcher constructed
    # below to use a Signal to dispatch callbacks.
    from qtsass.watchers.qt import QApplication

    qt_app = QApplication.instance()
    if not qt_app:
        qt_app = QApplication([])

    watch_dir = tmpdir.join('src').strpath
    os.makedirs(watch_dir)
    shutil.copy2(example('dummy.scss'), watch_dir)
    input = tmpdir.join('src/dummy.scss').strpath
    output = tmpdir.join('build/dummy.css').strpath
    output_exists = lambda: exists(output)

    c = CallCounter()
    w = QtWatcher(
        watch_dir=watch_dir,
        compiler=compile_filename,
        args=(input, output),
    )
    # We connect a counter directly to the Watcher's Qt Signal in order to
    # verify that the Watcher is actually using a Qt Signal.
    w.qtdispatcher.signal.connect(c)
    w.start()

    touch(input)
    time.sleep(0.5)
    if not await_condition(output_exists, qt_app=qt_app):
        assert False, 'Output file not created...'
    assert c.count == 1

    # Stop watcher
    w.stop()
    w.join()


def test_retry():
    """Test retry decorator"""

    @retry(5, interval=0)
    def succeeds_after(n, counter):
        counter()
        if n <= counter.count:
            return True
        raise ValueError

    # Succeed when attempts < retries
    assert succeeds_after(4, CallCounter())

    # Fails when retries < attemps
    with pytest.raises(ValueError):
        assert succeeds_after(6, CallCounter())

    @retry(5, interval=0)
    def fails():
        raise ValueError

    # Most obvious case
    with pytest.raises(ValueError):
        fails()
